#! /usr/bin/env python
"""Skeleton project file mainloop for new OLPCGames users"""
import os, random, pygame, logging 
try:
  import olpcgames
except:
  pass

log = logging.getLogger( 'inarow run' )
log.setLevel( logging.DEBUG )

screen = None
dirtyRects = []

# class to create static functions
class Callable:
  def __init__(self, func):
    self.__call__ = func

class Button:
  def __init__(self, name, bStr=None, bHiliteStr=None, bPressedStr=None, bgPic=None, r=None):
    self.name = name
    self.buttonUp = self.load(bStr)    
    self.buttonHilite = self.load(bHiliteStr)
    self.buttonPressed = self.load(bPressedStr)    
    self.button = self.buttonUp
    self.bgPic = bgPic
    self.r = r
    self.x = 0;
    self.y = 0
    self.isEnabled = True
    self.isHidden = False

  def setXY(self, x, y):
    self.x = x
    self.y = y
    self.r.x += x
    self.r.y += y

  def enable(self, isEnabled):
    self.isEnabled = isEnabled
    self.button = self.buttonUp

    self.draw()

  def hide(self, isHidden):
    self.isHidden = isHidden
    if isHidden:
      self.enable(False)

  def load(self, str):
    ret = None
    if str != None:
      ret = pygame.image.load(os.path.join("data", str)) # .convert_alpha()

    return ret

  def dup(self, name, x=0, y=0):
    b = Button(name)
    b.buttonUp = self.buttonUp
    b.buttonHilite = self.buttonHilite
    b.buttonPressed = self.buttonPressed
    b.bgPic = self.bgPic
    b.r = pygame.Rect(self.r)
    b.setXY(x, y)
    b.button = b.buttonUp

    return b

  def pointOn(self, pos):
    # check to see if point in the button update area
    return self.r.collidepoint(pos[0], pos[1])

  def handleMouseMove(self, pos):
    if not self.isEnabled:
      return
    isPointing = self.pointOn(pos)
    if isPointing and self.button != self.buttonHilite:
      self.button = self.buttonHilite
      self.draw()
    elif not isPointing and self.button != self.buttonUp:
      self.button = self.buttonUp
      self.draw()

  def handleClick(self, pos):
    if self.isEnabled and self.pointOn(pos):
      self.doClick()
      return self.name

    return None

  def doClick(self):
    self.button = self.buttonPressed
    self.draw()
    pygame.display.update(self.r)

  def eraseButtonArea(self):
    screen.set_clip(self.r)
    screen.blit(self.bgPic, (0, 0))
    screen.set_clip(screen.get_rect())

  def draw(self):
    if not self.isHidden:
      if not self.isEnabled:
        self.button.set_alpha(50)
        self.eraseButtonArea()
      screen.blit(self.button, (self.x, self.y))
      self.button.set_alpha(None)
      dirtyRects.append(self.r)

class StateButton(Button):
  def __init__(self, name, stateStrs=None, bHiliteStr=None, bgPic=None, r=None):
    self.name = name
    self.stateButtons = []
    self.loadStates(stateStrs)
    self.buttonHilite = self.load(bHiliteStr)
    self.bgPic = bgPic
    self.x = 0
    self.y = 0
    self.r = r
    self.isEnabled = True
    self.isHidden = False
    self.isHiliteOn = False

    self.currButton = 0
    if len(self.stateButtons) > 0:
      self.button = self.stateButtons[0]
  
  def loadStates(self, stateStrs):
    if stateStrs == None:
      return
    for str in stateStrs:
      self.stateButtons.append(self.load(str))

  def getState(self):
    return self.currButton

  def setState(self, index):
    self.currButton = index
    if self.currButton >= len(self.stateButtons):
      self.currButton = 0
      
    self.button = self.stateButtons[self.currButton]

  def enable(self, isEnabled):
    self.isEnabled = isEnabled

    self.draw()

  def nextState(self):
    self.setState(self.currButton + 1)

  def dup(self, name, x=0, y=0):
    b = StateButton(name)
    b.stateButtons = self.stateButtons
    b.buttonHilite = self.buttonHilite
    b.bgPic = self.bgPic
    b.r = pygame.Rect(self.r)
    b.setXY(x, y)
    b.button = b.stateButtons[0]

    return b

  def drawHilite(self):
    screen.blit(self.buttonHilite, (self.x, self.y))

  def handleMouseMove(self, pos):
    if not self.isEnabled:
      return
    isPointing = self.pointOn(pos)
    if isPointing and not self.isHiliteOn:
      self.isHiliteOn = True
      self.draw()
      self.drawHilite()
    elif not isPointing and self.isHiliteOn:
      self.isHiliteOn = False
      self.eraseButtonArea()
      self.draw()

  def doClick(self):
    self.nextState()
    self.draw()

class ButtonManager:
  def __init__(self):
    self.buttons = []

  def append(self, button):
    self.buttons.append(button)

  def enable(self, name, isEnabled):
    for button in self.buttons:
      if button.name == name:
        button.enable(isEnabled)

  def hide(self, name, isHidden):
    for button in self.buttons:
      if button.name == name:
        button.hide(isHidden)

  def get(self, name):
    for button in self.buttons:
      if button.name == name:
        return button

  def draw(self):
    for button in self.buttons:
      button.draw()

  def handleMouseMove(self, pos):
    for button in self.buttons:
      button.handleMouseMove(pos)

  def handleClick(self, pos):
    for button in self.buttons:
      clicked = button.handleClick(pos)
      if clicked != None:
        return clicked

class UI:
  X_TYPE = 0
  O_TYPE = 1
  WIDTH = 300

  def __init__(self, screenSize):
    self.screenSize = screenSize
    self.playerTypes = [ Player.LOCAL_PERSON, Player.WEAK_COMPUTER ]
    self.buttonsEnabled = True
    self.load()
    self.draw()

    self.gameRect = pygame.Rect(UI.WIDTH, 0, screenSize[0] - UI.WIDTH, screenSize[1])
    self.rows = 3
    self.columns = 3
    self.numToWin = 3
    self.drawNumber()
    self.createGame()

  def createGame(self):
    # erase old grid
    screen.set_clip(self.gameRect)
    screen.blit(self.backgroundPic, (0, 0))
    screen.set_clip(screen.get_rect())
    self.game = GridGame(self.gameRect, self.rows, self.columns, self.numToWin)
    self.game.setPlayers(self.playerTypes)
    self.game.drawGrid()
    self.game.startTurn()
    self.game.update()
    self.setButtonsEnableCorrectly(True)

  def loadNumbers(self):
    self.numberPic = [ None, None, None ]
    # load numbers 3 to 7 inclusive
    for i in range(3, 8):
      self.numberPic.append(pygame.image.load(os.path.join("data", "%d.gif" % i)))

  def drawNumber(self):
    screen.blit(self.numberPic[self.numToWin], (0, 0))
    dirtyRects.append(pygame.Rect(48, 46, 82, 87))

  def load(self):
    self.loadNumbers()
    self.backgroundPic = pygame.image.load(os.path.join("data", "bg.png"))
    self.bezelPic = pygame.image.load(os.path.join("data", "bezel.gif")).convert()

    self.upButton = Button("horizup", "up.gif", "up_hilite.gif", "up_pressed.gif", \
                           self.bezelPic, pygame.Rect(215, 148, 67, 52))
    self.downButton = Button("horizdown", "down.gif", "down_hilite.gif", "down_pressed.gif", \
                           self.bezelPic, pygame.Rect(215, 200, 67, 52))
    self.okButton = Button("ok", "ok.gif", "ok_hilite.gif", "ok_pressed.gif", \
                           self.bezelPic, pygame.Rect(51, 698, 220, 81))
    self.stopButton = Button("stop", "stop.gif", "stop_hilite.gif", "stop_pressed.gif", \
                           self.bezelPic, pygame.Rect(51, 698, 220, 81))
    self.stopButton.hide(True)

    playerStrs = [ "player_computer1.gif", "player_computer2.gif", \
                   "player_computer3.gif", "player_laptop.png", "player_mesh.png" ]
    self.playerButton = StateButton("xplayer", playerStrs, "player_hilite.png", \
                           self.bezelPic, pygame.Rect(120, 410, 170, 133))

    self.bm = ButtonManager()
    self.bm.append(self.upButton)
    self.bm.append(self.downButton)
    self.bm.append(self.upButton.dup("vertup", 0, 140))
    self.bm.append(self.downButton.dup("vertdown", 0, 140))
    self.bm.append(self.upButton.dup("inarowup", 0, -110))
    self.bm.append(self.downButton.dup("inarowdown", 0, -110))
    self.bm.append(self.playerButton)
    self.bm.append(self.playerButton.dup("oplayer", 0, 140))
    self.bm.append(self.okButton)
    self.bm.append(self.stopButton)

    self.bm.get("xplayer").setState(self.playerTypes[UI.X_TYPE])
    self.bm.get("oplayer").setState(self.playerTypes[UI.O_TYPE])

  def draw(self):
    screen.blit(self.backgroundPic, (0, 0))
    screen.blit(self.bezelPic, (0, 0))
    self.bm.draw()
    dirtyRects.append(((0, 0), self.screenSize))

  def setButtonsEnableCorrectly(self, globalIsEnabled):
    self.bm.enable("inarowup", globalIsEnabled)
    self.bm.enable("inarowdown", globalIsEnabled)
    if self.numToWin >= self.rows or self.numToWin >= self.columns or self.numToWin >= 7:
      self.bm.enable("inarowup", False)
    if self.numToWin <= 3:
      self.bm.enable("inarowdown", False)

    self.bm.enable("horizup", globalIsEnabled)
    self.bm.enable("horizdown", globalIsEnabled)
    if self.rows >= 50:
      self.bm.enable("horizup", False)
    if self.rows <= 3:
      self.bm.enable("horizdown", False)
      
    self.bm.enable("vertup", globalIsEnabled)
    self.bm.enable("vertdown", globalIsEnabled)
    if self.columns >= 50:
      self.bm.enable("vertup", False)
    if self.columns <= 3:
      self.bm.enable("vertdown", False)
      
    self.bm.enable("xplayer", globalIsEnabled)
    self.bm.enable("oplayer", globalIsEnabled)

    self.bm.hide("ok", not globalIsEnabled)
    self.bm.enable("ok", globalIsEnabled)
    self.bm.hide("stop", globalIsEnabled)
    self.bm.enable("stop", not globalIsEnabled)
      
    self.buttonsEnabled = globalIsEnabled

  def handleMouseMove(self, pos):
    self.game.handleMouseMove(pos)
    self.bm.handleMouseMove(pos)

  def handleClick(self, pos):
    started = self.game.handleClick(pos)
    clicked = self.bm.handleClick(pos)
    if clicked == "inarowup":
      self.numToWin += 1
      self.drawNumber()
      self.createGame()
    elif clicked == "inarowdown":
      self.numToWin -= 1
      self.drawNumber()
      self.createGame()
    elif clicked == "horizup":
      self.rows += 1
      self.createGame()
    elif clicked == "horizdown":
      self.rows -= 1
      self.createGame()
    elif clicked == "vertup":
      self.columns += 1
      self.createGame()
    elif clicked == "vertdown":
      self.columns -= 1
      self.createGame()
    elif clicked == "xplayer":
      self.playerTypes[UI.X_TYPE] = self.bm.get("xplayer").getState()  
      self.createGame()
    elif clicked == "oplayer":
      self.playerTypes[UI.O_TYPE] = self.bm.get("oplayer").getState()  
      self.createGame()
    elif clicked == "ok":
      self.createGame()
      self.setButtonsEnableCorrectly(False)
      self.game.start()
    elif clicked == "stop":
      self.createGame()
      self.setButtonsEnableCorrectly(True)
    elif started:
      self.setButtonsEnableCorrectly(False)

  def update(self):
    done = self.game.update()
    if done:
      self.setButtonsEnableCorrectly(True)

class GridCell:
  def __init__(self, r, owner):
    self.r = r
    self.owner = owner
    self.hover = False
    self.saveBehind = None

  def pointIn(self, pos):
    return self.r.collidepoint(pos)

  def centerPoint(self):
    return [ self.r.left + self.r.width / 2, self.r.top + self.r.height / 2 ]

  def draw(self):
    if (self.owner == Player.NONE):
      return ()

    if self.owner.token == Player.X:
      w = int(self.r.width / 7)
      pad = int(self.r.width * 0.2)

      start = (self.r.left + pad, self.r.top + pad)
      end = (self.r.left + self.r.width - pad, self.r.top + self.r.height - pad)
      pygame.draw.line(screen, (255, 255, 255), start, end, w)

      start = (self.r.left + self.r.width - pad, self.r.top + pad)
      end = (self.r.left + pad, self.r.top + self.r.height - pad)
      pygame.draw.line(screen, (255, 255, 255), start, end, w)
    else:
      w = int(self.r.width / 9)
      left = int(self.r.left + self.r.width / 2)
      top = int(self.r.top + self.r.height / 2)
      radius = int(self.r.width * 0.3)
      pygame.draw.circle(screen, (255, 255, 255), (left, top), radius, w)
      # do another circle next to it since the circle() function can leave gaps
      pygame.draw.circle(screen, (255, 255, 255), (left + 1, top), radius, w)

    return self.r

  def set(self, player):
    self.owner = player
    self.setHover(False)
    return self.draw()

  def drawHover(self, change):
    r = None
    if self.hover and change:
      # make a child surface with the background before hover
      s = screen.subsurface(self.r)
      # make a place to hold the background
      self.saveBehind = pygame.Surface((self.r.width, self.r.height), 0, screen)
      # make a copy of it on its own surface
      self.saveBehind.blit(s, (0, 0))

      # draw alpha hilite
      h = pygame.Surface((self.r.width, self.r.height), 0, screen)
      h.fill((255, 255, 55))
      h.set_alpha(70)
      screen.blit(h, self.r)
      r = self.r
    elif not self.hover and self.saveBehind != None:
      screen.blit(self.saveBehind, self.r)
      self.saveBehind = None
      r = self.r

    return r
    
  def getHover(self):
    return self.hover

  def setHover(self, value):
    change = (self.hover != value)
      
    self.hover = value
    return self.drawHover(change)

class GridGame:
  def __init__(self, gameRect, rows, columns, numToWin):
    self.gameRect = gameRect
    self.rows = rows
    self.columns = columns
    self.numToWin = numToWin
    self.borderPadding = 50
    self.lineColor = (250, 150, 150)

    self.players = []
    self.turn = 0
    self.started = False
    self.done = False
    self.winner = Player.NONE

    self.makeGrid()

  def start(self):
    ret = not self.started
    self.started = True

    return ret

  def setPlayers(self, playerTypes):
    self.players = []
    self.players.append(Player.create(playerTypes[UI.X_TYPE], "X", Player.X))
    self.players.append(Player.create(playerTypes[UI.O_TYPE], "O", Player.O))

  def makeGrid(self):
    pxPerColumn = (self.gameRect.width - self.borderPadding) / self.columns
    pxPerRow = (self.gameRect.height - self.borderPadding) / self.rows
    self.cellWH = min(pxPerColumn, pxPerRow)
    self.lineW = int(self.cellWH / 10) + 1
    x = self.gameRect.x + (self.gameRect.width - self.cellWH * self.columns) / 2
    y = self.gameRect.y + (self.gameRect.height - self.cellWH * self.rows) / 2
    self.grid = []
    for r in range(self.rows):
      row = []
      for c in range(self.columns):
        rect = pygame.Rect(x + c * self.cellWH, y + r * self.cellWH, \
                           self.cellWH, self.cellWH)
        cell = GridCell(rect, 0)
        row.append(cell)
      self.grid.append(row)

  def drawGrid(self):
    # draw horizontal lines for rows
    x = self.grid[0][0].r.left
    y = self.grid[0][0].r.top - int(self.lineW / 2)
    width = self.cellWH * self.columns
    height =  self.lineW
    for r in range(self.rows):
      if r != 0:
        screen.fill(self.lineColor, (x, y, width, height))
      y += self.cellWH

    # draw vertical lines for columns
    x = self.grid[0][0].r.left - int(self.lineW / 2)
    y = self.grid[0][0].r.top
    width = self.lineW
    height = self.cellWH * self.rows
    for c in range(self.columns):
      if c != 0:
        screen.fill(self.lineColor, (x, y, width, height))
      x += self.cellWH

    dirtyRects.append(self.gameRect)

  def startTurn(self):
    self.players[self.turn].startTurn(self)
    self.handleMouseMove(pygame.mouse.get_pos())

  def turnDone(self):
    self.turn += 1
    if self.turn >= len(self.players):
      self.turn = 0

    self.startTurn()

  def inGrid(self, r, c):
    return 0 <= r and r < self.rows and 0 <= c and c < self.columns

  def listEndsOpen(self, lst):
    if len(lst) < 2:
      return False

    # find the direction of the list in delta_c and delta_r
    dr = lst[1][0] - lst[0][0]
    dc = lst[1][1] - lst[0][1]

    r1 = lst[0][0] - dr
    c1 = lst[0][1] - dc
    r2 = lst[-1][0] + dr
    c2 = lst[-1][1] + dc

    if self.inGrid(r1, c1) and self.grid[r1][c1].owner == Player.NONE \
            and self.inGrid(r2, c2) and self.grid[r2][c2].owner == Player.NONE:
      return True

    return False

  def getInARowOneDir(self, r, c, dr, dc, owner):
    ret = []
    while self.inGrid(r + dr, c + dc) and \
          self.grid[r + dr][c + dc].owner != Player.NONE and \
          owner.token == self.grid[r + dr][c + dc].owner.token:
      r += dr
      c += dc
      ret.append( ( r, c ) )

    return ret

  def getInARow(self, r, c, dr, dc, simulateWith=None):
    owner = self.grid[r][c].owner
    if simulateWith != None:
      self.grid[r][c].owner = simulateWith
    elif owner == Player.NONE:
      return []

    forwardLst = self.getInARowOneDir(r, c, dr, dc, self.grid[r][c].owner)
    ret = self.getInARowOneDir(r, c, -dr, -dc, self.grid[r][c].owner)

    ret.reverse()
    ret.append( ( r, c ) )
    ret.extend(forwardLst)

    if simulateWith != None:
      self.grid[r][c].owner = owner

    return ret

  def getAnyDirInARow(self, r, c, simulateWith=None):
    # check the vertical line
    lst = self.getInARow(r, c, 1, 0, simulateWith)
    # check the horizontal line
    tmp = self.getInARow(r, c, 0, 1, simulateWith)
    if len(tmp) > len(lst) or len(tmp) == len(lst) and self.listEndsOpen(tmp):
      lst = tmp
    # check the diagonal line
    tmp = self.getInARow(r, c, 1, 1, simulateWith)
    if len(tmp) > len(lst) or len(tmp) == len(lst) and self.listEndsOpen(tmp):
      lst = tmp
    # check the other diagonal line
    tmp = self.getInARow(r, c, 1, -1, simulateWith)
    if len(tmp) > len(lst) or len(tmp) == len(lst) and self.listEndsOpen(tmp):
      lst = tmp

    return lst

  def getAnyDirInARowForOpponents(self, r, c, player):
    longest = []
    for opponent in self.players:
      if opponent != player:
        lst = self.getAnyDirInARow(r, c, opponent)
        if len(lst) > len(longest):
          longest = lst

    return longest

  def winningList(self):
    for r in range(self.rows):
      for c in range(self.columns):
        if self.grid[r][c].owner != Player.NONE:
          lst = self.getAnyDirInARow(r, c)
          if len(lst) >= self.numToWin:
            self.winner = self.grid[r][c].owner.token
            return lst

    return None

  def drawWinningLine(self, lst):
    start = lst[0]
    end = lst[-1]

    # set the line to end in the middle of the cells
    r, c = start
    startpos = self.grid[r][c].centerPoint()
    r, c = end
    endpos = self.grid[r][c].centerPoint()

    # depending on what line, make it longer by adjusting the endpoints away from the middle
    if start[0] == end[0]:
      # drawing horizontal line
      startpos[0] -= int(self.cellWH / 3)
      endpos[0] += int(self.cellWH / 3)
    elif start[1] == end[1]:
      # drawing vertical line
      startpos[1] -= int(self.cellWH / 3)
      endpos[1] += int(self.cellWH / 3)
    elif start[1] < end[1]:
      # drawing diagonal line
      startpos[0] -= int(self.cellWH / 3)
      startpos[1] -= int(self.cellWH / 3)
      endpos[0] += int(self.cellWH / 3)
      endpos[1] += int(self.cellWH / 3)
    else:
      # drawing the other diagonal line
      startpos[0] += int(self.cellWH / 3)
      startpos[1] -= int(self.cellWH / 3)
      endpos[0] -= int(self.cellWH / 3)
      endpos[1] += int(self.cellWH / 3)

    w = int(self.cellWH * 0.1) + 2

    rect = pygame.draw.line(screen, (200, 50, 80), startpos, endpos, w)

    dirtyRects.append(rect)

  def update(self):
    if not self.started:
      return False
    if not self.done:
      self.players[self.turn].checkReady(self)
    
    lst = self.winningList() 
    if lst != None:
      self.drawWinningLine(lst) 
    else:
      for r in range(self.rows):
        for c in range(self.columns):
          if self.grid[r][c].owner == Player.NONE:
            return False;

    self.done = True
    return True

  def selectCell(self, cell, player):
    r = cell.set(player)
    self.turnDone()
    dirtyRects.append(r)

  def handleMouseMove(self, pos):
    if not self.done and isinstance(self.players[self.turn], LocalPlayer):
      # look to see what cell (if any) is hovered over
      for row in self.grid:
        for cell in row:
          if cell.pointIn(pos) and cell.owner == Player.NONE:
            dirtyRects.append(cell.setHover(True))
          elif cell.getHover():
            dirtyRects.append(cell.setHover(False))

  def handleClick(self, pos):
    if not self.done and isinstance(self.players[self.turn], LocalPlayer):
      # look to see what cell (if any) was clicked
      for row in self.grid:
        for cell in row:
          if cell.pointIn(pos) and cell.owner == Player.NONE:
            self.players[self.turn].clickedCell(cell)
            return self.start()

    return False

class Player:
  NONE = 0
  X = 1
  O = 2

  WEAK_COMPUTER = 0
  DEFENSIVE_COMPUTER = 1
  OFFENSIVE_COMPUTER = 2
  LOCAL_PERSON = 3
  MESH_PERSON = 4
  NUM_TYPES = 5

  def __init__(self, name, token):
    self.name = name
    self.token = token
    self.myTurn = False
    self.turns = 0
    self.cellChoice = None

  def startTurn(self, game):
    self.turns += 1

  def checkReady(self, game):
    if self.cellChoice != None:
      game.selectCell(self.cellChoice, self)
      self.cellChoice = None

  def create(type, name, token):
    if type == Player.WEAK_COMPUTER:
      return WeakComputerPlayer(name, token)
    elif type == Player.DEFENSIVE_COMPUTER:
      return DefensiveComputerPlayer(name, token)
    elif type == Player.OFFENSIVE_COMPUTER:
      return OffensiveComputerPlayer(name, token)
    elif type == Player.LOCAL_PERSON:
      return LocalPlayer(name, token)
    elif type == Player.MESH_PERSON:
      return MeshPlayer(name, token)
  create = Callable(create)

class LocalPlayer(Player):
  def __init__(self, name, token):
    Player.__init__(self, name, token)

  def clickedCell(self, cell):
    self.cellChoice = cell

class MeshPlayer(Player): 
  def __init__(self, name, token):
    Player.__init__(self, name, token)

class ComputerPlayer(Player):
  def __init__(self, name, token):
    Player.__init__(self, name, token)
    self.matrix = None
    self.saveTillTicks = 0

  def startTurn(self, game):
    Player.startTurn(self, game)
    # add half sec pause for computer "thinking"
    self.saveTillTicks = pygame.time.get_ticks() + 500

  def checkReady(self, game):
    if self.saveTillTicks != 0 and self.saveTillTicks < pygame.time.get_ticks():
      self.weightCells(game)
      row, column = self.pickBestCell()
      self.cellChoice = game.grid[row][column]
      self.saveTillTicks = 0

    Player.checkReady(self, game)

  def weightCells(self, game):
    self.initMatrix(game)

  def createMatrix(self, game):
    self.matrix = []
    for r in range(game.rows):
      row = []
      for c in range(game.columns):
        row.append(0)
      self.matrix.append(row)

  def initMatrix(self, game):
    if self.matrix == None or game.rows != len(self.matrix) \
                           or game.columns != len(self.matrix[0]):
      self.createMatrix(game)

    for r in range(game.rows):
      for c in range(game.columns):
        self.matrix[r][c] = 0

  def pickBestCell(self):
    high = 0
    lst = []
    
    for r in range(len(self.matrix)):
      for c in range(len(self.matrix[r])):
        if self.matrix[r][c] > high:
          high = self.matrix[r][c]
          lst = [ ( r, c ) ]
        elif self.matrix[r][c] == high:
          lst.append( ( r, c ) )

    rand = random.randrange(0, len(lst))

    return lst[rand]
  
class WeakComputerPlayer(ComputerPlayer): 
  def __init__(self, name, token):
    ComputerPlayer.__init__(self, name, token)

  def weightCells(self, game):
    ComputerPlayer.weightCells(self, game)

    for r in range(game.rows):
      for c in range(game.columns):
        if game.grid[r][c].owner == Player.NONE:
          # check to see if can win
          lst = game.getAnyDirInARow(r, c, self)
          if len(lst) >= game.numToWin:
            self.matrix[r][c] += 100
          
          # see where opponent is getting close
          lst = game.getAnyDirInARowForOpponents(r, c, self)
          if len(lst) > game.numToWin / 2:
            self.matrix[r][c] += 10
          elif len(lst) > 1:
            self.matrix[r][c] += 5
  
class DefensiveComputerPlayer(ComputerPlayer):
  def __init__(self, name, token):
    ComputerPlayer.__init__(self, name, token)

  def weightCells(self, game):
    ComputerPlayer.weightCells(self, game)

    for r in range(game.rows):
      for c in range(game.columns):
        if game.grid[r][c].owner == Player.NONE:
          # check to see if can win
          lst = game.getAnyDirInARow(r, c, self)
          if len(lst) >= game.numToWin:
            self.matrix[r][c] += 100
          
          # see where opponent has the longest run
          lst = game.getAnyDirInARowForOpponents(r, c, self)
          self.matrix[r][c] += 10 * len(lst)
          # if the ends of the list are open it's better
          if game.listEndsOpen(lst):
            self.matrix[r][c] += 5 
          # give slight preference to a diagonal line
          if len(lst) >= 2 and lst[0][0] != lst[1][0] and lst[0][1] != lst[1][1]:
            self.matrix[r][c] += 1
          # in tic tac toe type give preference to corners
          if game.rows == game.columns and game.rows == game.numToWin \
                          and (r == 0 or r == game.rows - 1) \
                          and (c == 0 or c == game.columns - 1):
            self.matrix[r][c] += 1
            
 
class OffensiveComputerPlayer(ComputerPlayer): 
  def __init__(self, name, token):
    ComputerPlayer.__init__(self, name, token)
  
  def weightCells(self, game):
    ComputerPlayer.weightCells(self, game)

    for r in range(game.rows):
      for c in range(game.columns):
        if game.grid[r][c].owner == Player.NONE:
          # mlst is my list (the offensive computer)
          # tlst is their list (my opponent)
          mlst = game.getAnyDirInARow(r, c, self)
          tlst = game.getAnyDirInARowForOpponents(r, c, self)

          # check to see if can win
          if len(mlst) >= game.numToWin:
            self.matrix[r][c] += 50000

          # check to see if can lose
          elif len(tlst) >= game.numToWin:
            self.matrix[r][c] += 20000

          # check to see if can set up a win
          elif len(mlst) == game.numToWin - 1 and game.listEndsOpen(mlst):
            self.matrix[r][c] += 10000

          # check to see if oppenent can set up a win
          elif len(tlst) == game.numToWin - 1 and game.listEndsOpen(tlst):
            self.matrix[r][c] += 5000

          # take my best pick
          else:
            self.matrix[r][c] += 8 * len(mlst)
            self.matrix[r][c] += 10 * len(tlst)

          # if the ends of the list are open it's better
          if game.listEndsOpen(tlst):
            self.matrix[r][c] += 5 
          # give slight preference to a diagonal line
          if len(tlst) >= 2 and tlst[0][0] != tlst[1][0] and tlst[0][1] != tlst[1][1]:
            self.matrix[r][c] += 1
          # in tic tac toe type give preference to corners
          if game.rows == game.columns and game.rows == game.numToWin \
                          and (r == 0 or r == game.rows - 1) \
                          and (c == 0 or c == game.columns - 1):
            self.matrix[r][c] += 1

def main():
  global screen, dirtyRects

  try:
    size = olpcgames.ACTIVITY.game_size
  except:
    size = (1200,900-75)

  screen = pygame.display.set_mode(size)
  ui = UI(size)
  clock = pygame.time.Clock()

  running = True
  while running:
    events = pygame.event.get()

    for event in events:
      if event.type == pygame.MOUSEMOTION:
        ui.handleMouseMove(event.pos)
      elif event.type == pygame.QUIT:
        running = False
      elif event.type == pygame.MOUSEBUTTONDOWN:
        ui.handleClick(event.pos)
      elif event.type == pygame.KEYDOWN:
        if event.key == pygame.K_ESCAPE:
          running = False
        elif event.key == pygame.K_s:
          pygame.display.flip()

    ui.update()

    pygame.display.update(dirtyRects)
    dirtyRects = []

    clock.tick(10) # maximum number of frames per second
        

if __name__ == "__main__":
  logging.basicConfig()
  main()
